"""Settings module"""
import configparser
import logging
import logging.config
import os
import shutil

CONFIG_INI = None
NOTIFY = None

NOTIFY = True
PATH_TEMPLATES = (
    # '/tmp/{home}/.config/swytcher/{filename}',
    '{home}/.config/swytcher/{filename}',
    # '{home}/.local/swytcher/config/{filename}',
)


log = logging.getLogger(__name__)  # pylint: disable=invalid-name


def _setup_logging(config: configparser.ConfigParser) -> None:
    # Setup logging
    log_conf = 'log_conf.ini'
    logfile = get_config(log_conf)

    if not logfile:  # pragma: no cover; how to mock a configparser object?
        logfile = conf_not_found(log_conf)

    logging.config.fileConfig(logfile, disable_existing_loggers=False)

    # Optionally override logging from config.ini
    loglevel = config['logging'].get('loglevel')
    if loglevel:
        logger = logging.getLogger()
        logger.setLevel(logging.getLevelName(loglevel))  # type: ignore
        log.info("Loglevel %r set!", loglevel)


def _make_conf(config_ini: str) -> str:
    config_file = conf_not_found(
        config_ini, log_msg=True, config_paths=conf_paths(config_ini))
    cp_conf_path = conf_paths(config_ini)[0]
    log.info("Copying default conf to %r", cp_conf_path)
    os.makedirs(os.path.dirname(cp_conf_path), exist_ok=True)
    config_file = shutil.copy(config_file, cp_conf_path)
    return config_file


def _get_configparser() -> configparser.ConfigParser:
    config = configparser.ConfigParser()
    config_ini = 'config.ini'
    config_file = get_config(config_ini) or _make_conf(config_ini)
    log.info("Using config found in %r", config_file)
    config.read(config_file)
    if not config:
        raise FileNotFoundError("{} not found".format(config_ini))
    return config


def _setup_config(config: configparser.ConfigParser) -> None:
    # Setup globals from config.ini
    global CONFIG_INI  # pylint: disable=global-statement
    CONFIG_INI = config

    global NOTIFY  # pylint: disable=global-statement
    NOTIFY = config['logging'].getboolean('notify')  # type: ignore


def load_configs() -> None:
    """Load configs"""
    config = _get_configparser()
    _setup_config(config)
    _setup_logging(config)


def setup_layouts(xkb, config):
    """Setup layout settings"""
    if not config:
        raise AssertionError("Config file must be loaded first")

    layout_sections = [
        (name, section) for name, section in config.items()
        if name.startswith('layout')
    ]

    layouts = []
    for i, (name, section) in enumerate(layout_sections):
        log.info("Config section %r used for %r", name, xkb.groups_names[i])
        layouts.append({
            'name': xkb.groups_names[i],
            'strings': section['strings'].strip().splitlines(),
            'substrings': section['substrings'].strip().splitlines(),
        })

    return layouts


def conf_paths(filename) -> list:
    """Get config paths"""
    home = os.path.expanduser('~')
    paths = [path.format(home=home, filename=filename) for path in
             PATH_TEMPLATES]
    return paths


def get_config(filename: str) -> str:
    """Try to find user configured logfile"""
    config_file = ''
    for path in conf_paths(filename):
        if os.path.isfile(path):
            config_file = path
            break
    return config_file


def conf_not_found(
        filename: str, log_msg: bool = False,
        config_paths: list = None) -> str:
    """Log warning that config file was not found, return path to default
    conf"""
    default_conf = "%s%s%s" % (os.path.dirname(__file__), os.path.sep,
                               filename)

    if log_msg:
        log.warning(
            "Config file %r not found%s; using default %r",
            filename,
            " in %s" % ' or '.join(config_paths) if config_paths else "",
            default_conf)
    return default_conf



# -*- coding: utf-8 -*-
"""CLI frontend for Swytcher"""
import logging

import click

import swytcher.settings as settings
import swytcher.swytcher as swytcher

log = logging.getLogger(__name__)  # pylint: disable=invalid-name
logging.basicConfig(level=logging.INFO)


@click.command()
def main(args=None):
    """Console script for swytcher"""
    # click.echo("See click documentation at http://click.pocoo.org/")
    click.echo(__name__)
    settings.load_configs()  # pylint: disable=protected-access

    swytcher.main(args)


if __name__ == "__main__":  # pragma: no cover
    main()



"""Tests for settings.py"""
# pylint: disable=invalid-name
# pylint: disable=missing-docstring
# pylint: disable=protected-access
import collections
import unittest

import pytest

import swytcher.settings as settings


@pytest.fixture
def os_mock(monkeypatch):
    """Path out os"""
    mock_os = unittest.mock.MagicMock(spec=settings.os)
    mock_os.path.expanduser.return_value = '/home/foouser'
    monkeypatch.setattr(settings, 'os', mock_os)
    return mock_os


# setup_layouts
def test_setup_layouts():
    """Test setup_layouts"""
    class TestXkb:  # pylint: disable=too-few-public-methods
        groups_names = ('layout1', 'layout2')

    config = collections.OrderedDict((
        ('layout_1', {'strings': 'foo', 'substrings': 'bar'}),
        ('layout_2', {'strings': 'baz', 'substrings': 'qux'}),
    ))
    msettings = settings.setup_layouts(TestXkb(), config=config)
    expected = [
        {'name': 'layout1', 'strings': ['foo'], 'substrings': ['bar']},
        {'name': 'layout2', 'strings': ['baz'], 'substrings': ['qux']},
    ]
    assert msettings == expected


def test_setup_layouts_config_not_loaded():
    """Test assertion error when config is not available"""
    with pytest.raises(AssertionError):
        class TestXkb:  # pylint: disable=too-few-public-methods
            groups_names = ('layout1', 'layout2')
        settings.setup_layouts(TestXkb(), None)


# _setup_logging
def test_setup_logging_conf_lacks_log_setting():
    """Test behavior when loglevel cannot be retrieved from config; this may
    happen if the config lacks a logging setting, which is OK."""
    config = {'logging': {}}
    settings._setup_logging(config)


# _get_configparser
def test_get_configparser_filenotfound(monkeypatch):
    mock_parser = unittest.mock.MagicMock(spec=settings.configparser)
    mock_parser.ConfigParser().__bool__.return_value = False
    monkeypatch.setattr(settings, 'configparser', mock_parser)
    with pytest.raises(FileNotFoundError):
        settings._get_configparser()


def test_get_configparser_default_conf(monkeypatch):
    mock_parser = unittest.mock.MagicMock(spec=settings.configparser)
    mock_parser.ConfigParser().__bool__.return_value = True
    monkeypatch.setattr(settings, 'configparser', mock_parser)

    get_config_mock = unittest.mock.MagicMock(
        spec=settings.get_config, return_value=None)
    monkeypatch.setattr(settings, 'get_config', get_config_mock)

    monkeypatch.setattr(settings.shutil, 'copy', unittest.mock.MagicMock)

    assert settings._get_configparser()


# get_config
def test_get_config(os_mock):
    # pylint: disable=redefined-outer-name
    """Test that get_config returns a string"""
    os_mock.path.isfile.return_value = True
    assert '/conf.ini' in settings.get_config('conf.ini')


def test_get_config_fail(os_mock):
    # pylint: disable=redefined-outer-name
    """Test that get_config returns a string"""
    os_mock.path.isfile.return_value = False
    assert settings.get_config('conf.ini') == ''


# conf_paths
def test_conf_paths(os_mock):
    # pylint: disable=redefined-outer-name,unused-argument
    assert (settings.conf_paths('foo_file')[0] ==
            '/home/foouser/.config/swytcher/foo_file')


# conf_not_found
def test_conf_not_found(os_mock):
    # pylint: disable=redefined-outer-name,unused-argument
    default_conf = settings.conf_not_found('name', ['path'])
    assert default_conf.endswith('name')



"""Automatic keyboard layout switcher"""
import functools
import logging
import subprocess
from typing import Iterable
from typing import Set

import xkbgroup

import swytcher.settings as settings
import swytcher.xwindow as xwindow
from swytcher.util import suppress_err

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


# Move this to swytcher.system
@suppress_err(FileNotFoundError, log)
def notify(title: str, msg: str = '') -> None:  # pragma: no cover
    """Use notify-send (if available) to inform user of layout switch."""
    if not settings.NOTIFY:
        return
    cmd = [
        'notify-send',
        '--urgency=low',
        '--expire-time=2000',
        title,
        msg
    ]
    subprocess.call(cmd)


def change_layout(xkb: xkbgroup.XKeyboard, layout: str) -> bool:
    """Set layout; returns True if layout was changed, False otherwise"""
    if xkb.group_name == layout:  # check against current layout
        log.debug("%r is already the active layout", layout)
        return False  # don't change layout if it's already correct
    log.info("setting layout %r", layout)
    xkb.group_name = layout
    notify("Changed layout", layout)
    return True


def _match_substrings(name_list: Iterable[str],
                      substrings: Iterable[str]) -> set:
    """Substring filter match"""
    found_matches = set()
    for name in name_list:
        for substring in substrings:
            if substring in name:
                log.debug("Substring filter match: %r in %r", substring, name)
                found_matches.update([name])

    return found_matches


def matches(name_list: Iterable[str], strings: Iterable[str],
            substrings: Iterable[str]) -> Set[str]:
    """Returns True if any of the strings in the two filters `strings` and
    `substrings` occur in `name_list`."""
    matched = (set(strings) & set(name_list) or
               _match_substrings(name_list, substrings or {}))
    if matched:
        log.debug('%r matched %r from %r or %r',
                  name_list, matched, strings, substrings)
    return matched


def change_callback(name_list, xkb, layouts: list) -> None:  # pragma: no cover
    """Event handler when active window is changed"""
    # NOTE: These extracted variables should be removed later
    primary_filter = layouts[0]['strings']
    primary_substrings = layouts[0]['substrings']
    primary = layouts[0]['name']
    secondary_filter = layouts[1]['strings']
    secondary_substrings = layouts[1]['substrings']
    secondary = layouts[1]['name']

    # matched_layout = match_layout(name_list, layouts)
    # if matched_layout:
    #   change_layout(xkb, matched_layout)
    # else:
    #   change_layout(xkb, last_remembered_layout_for_window)

    if matches(name_list, secondary_filter, secondary_substrings):
        change_layout(xkb, secondary)
    elif matches(name_list, primary_filter, primary_substrings):
        change_layout(xkb, primary)
    else:
        log.debug("%r: No match, using default layout", name_list)
        change_layout(xkb, xkb.groups_names[0])


def main(args=None):  # pragma: no cover
    """Main"""
    if not args:
        pass
    xkb = xkbgroup.XKeyboard()
    layouts = settings.setup_layouts(xkb, settings.CONFIG_INI)
    log.info("Layouts configured by setxkbmap: %s", layouts)
    partial_cb = functools.partial(change_callback, xkb=xkb, layouts=layouts)
    xwindow.run(partial_cb)



"""Util functions"""
import logging
from contextlib import wraps  # type: ignore
from typing import Type


def suppress_err(
        exception: Type[Exception], logger: logging.Logger=None,
        level=logging.WARNING, traceback=False):
    """Decorator that will catch `exception`, suppress it, and optionally log
    it with a traceback through the logger.

    Usage example:

        >>> @exception_handler(TypeError, log, "my msg")
            def func():
                return 'a' + 1

        >>> func()
        'WARNING [Can't convert 'int' object to str implicitly'] my msg

        E.g., it will be equivalent to implementing 'func' as:

        >>> def func():
                try:
                    return 'a' + 1
                except TypeError as err:
                    logging.error("You can't do that: %s", err)

        Multiple exceptions can be handled by just adding more decorators:

        >>> @exception_handler(TypeError, log, "You can't do that")
            @exception_handler(AttributeError, log, "You can't do that")
            def func():
                return 'a' + 1
    """
    def wrapper(func):
        """Wrapper for func"""
        @wraps(func)
        def inner(*args, **kwargs):
            """Inner function that actually calls func"""
            try:
                return func(*args, **kwargs)
            except exception as err:
                if logger:
                    logger.log(
                        level, "%r", err, exc_info=1 if traceback else 0)
        return inner
    return wrapper



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=redefined-outer-name
# pylint: disable=protected-access
"""
test_swytcher
----------------------------------

Tests for `swytcher` module.
"""
from unittest.mock import Mock

import pytest

from swytcher import swytcher


# Fixtures
@pytest.fixture(autouse=True)
def xkbgroup_patch(monkeypatch):
    """Patch out xkbgroup so that we don't accidentally change keyboard
    settings when running tests."""
    mock = Mock(spec=swytcher.xkbgroup)
    monkeypatch.setattr(swytcher, 'xkbgroup', mock)
    return mock


@pytest.fixture(autouse=True)
def notify_patch(monkeypatch):
    """Patch out the notify function so that we don't send notifications when
    running tests."""
    notify_mock = Mock()
    monkeypatch.setattr(swytcher, 'notify', notify_mock)
    return notify_mock


# change_layout
def test_change_layout_changed():
    """change_layout success"""
    xkb = Mock()
    xkb.group_name = 'original_layout'

    layout = 'new_layout'
    assert swytcher.change_layout(xkb, layout) is True
    assert xkb.group_name == layout


def test_change_layout_unchanged():
    """change_layout failure"""
    layout = 'current_layout'
    xkb = Mock()
    xkb.group_name = layout

    assert swytcher.change_layout(xkb, layout) is False
    assert xkb.group_name == layout


# matches
def test_matches_no_match():
    """matches should return False when nothing matches"""
    name_list = ('foo', 'spamm', 'baz')
    strings = ('spam', 'ham')
    substrings = ('eggs', 'parrot')
    result = swytcher.matches(name_list, strings, substrings)
    assert not result


def test_matches_strings():
    """should find string match"""
    name_list = ('foo', 'ham', 'baz')
    strings = ('spam', 'ham')
    substrings = ('eggs', 'parrot')
    result = swytcher.matches(name_list, strings, substrings)
    assert result == {'ham'}


def test_matches_sub_strings():
    """should find substring match"""
    name_list = ('foo', 'hamm', 'sausage eggs and bacon')
    strings = ('spam', 'ham')
    substrings = ('egg', 'parrot')
    result = swytcher.matches(name_list, strings, substrings)
    assert result == {'sausage eggs and bacon'}


# _match_substrings
def test_match_substrings_found():
    """should find substring match"""
    name_list = ('foo', 'bar', 'sausage eggs and bacon', 'baz')
    substrings = ('egg', 'parrot')
    result = swytcher._match_substrings(name_list, substrings)
    assert result == {'sausage eggs and bacon'}


def test_match_substrings_not_found():
    """should find substring match"""
    name_list = ('foo', 'bar', 'sausage and bacon', 'baz')
    substrings = ('egg', 'parrot')
    result = swytcher._match_substrings(name_list, substrings)
    assert not result



"""Tests for util.py"""
# pylint: disable=missing-docstring
import logging
import unittest.mock

import swytcher.util as util


def test_suppress_err():
    """happy path testing for exception handler"""
    @util.suppress_err(KeyError)
    @util.suppress_err(TypeError)
    @util.suppress_err(IndexError)
    def testfunc():
        return 'foo'
    assert testfunc() == 'foo'


def test_suppress_err_logs():
    """test that exception was logged after being suppressed"""
    logger = unittest.mock.MagicMock(spec=logging.getLogger(__name__))

    @util.suppress_err(IndexError, logger)
    def testfunc():
        raise IndexError("wrong")

    testfunc()
    assert logger.log.called


def test_suppress_err_no_logging():
    """test that exception was logged after being suppressed"""
    @util.suppress_err(IndexError)
    def testfunc():
        raise IndexError("wrong")

    testfunc()



"""Tests for cli.py"""
import unittest

import pytest
from click.testing import CliRunner

from swytcher import cli


@pytest.fixture(autouse=True)
def mock_swytcher(monkeypatch):
    """Patch out swytcher module"""
    mock = unittest.mock.MagicMock()
    monkeypatch.setattr(cli, 'swytcher', mock)
    return mock


def test_command_line_interface():
    """CLI should return help text"""
    runner = CliRunner()
    result = runner.invoke(cli.main)
    assert result.exit_code == 0
    assert 'swytcher.cli' in result.output
    help_result = runner.invoke(cli.main, ['--help'])
    assert help_result.exit_code == 0
    assert '--help  Show this message and exit.' in help_result.output



"""Tests for xwindow.py"""
# pylint: disable=redefined-outer-name
# pylint: disable=invalid-name
from unittest.mock import MagicMock

import pytest
import Xlib

from swytcher import xwindow


# fixtures
@pytest.fixture(autouse=True)
def xlib(monkeypatch):
    """Mock out Xlib"""
    mock_xlib = MagicMock(spec=xwindow.Xlib)
    monkeypatch.setattr(xwindow, 'Xlib', mock_xlib)
    return mock_xlib


# get_window_name
def test_get_window_name():
    """when get_wm_name() works"""
    window = MagicMock(spec=Xlib.xobject.drawable.Window)
    window_name = 'my name'
    window.get_wm_name.return_value = window_name
    name = xwindow.get_window_name(window)
    assert name == window_name


def test_get_window_name_closure():
    """when get_wm_name() doesn't work"""
    window = MagicMock(spec=Xlib.xobject.drawable.Window)
    window.get_wm_name.return_value = ''

    window_name = 'my name'
    window.get_full_text_property.return_value = window_name
    name = xwindow.get_window_name(window)
    assert name == window_name


def test_get_window_name_closer():
    """when get_wm_name() doesn't work"""
    window = MagicMock(spec=Xlib.xobject.drawable.Window)
    window.get_wm_name.return_value = ''
    window.get_full_text_property.return_value = ''

    name = xwindow.get_window_name(window)
    assert not name


# handle_xevent
def test_handle_xevent_type_not_valid():
    """Return false when event type is invalid"""
    event = MagicMock(spec=Xlib.X.PropertyNotify)
    event.type = 'invalid'
    result = xwindow.handle_xevent(event, callback=lambda *args: None)
    assert result is False


def test_handle_xevent_atom_not_valid():
    """Return false when event type is invalid"""
    event = MagicMock(spec=Xlib.protocol.event.PropertyNotify)
    event.type = xwindow.Xlib.X.PropertyNotify
    event.atom = 'invalid'
    result = xwindow.handle_xevent(event, callback=lambda *args: None)
    assert result is False


def test_handle_xevent_atom_ok(monkeypatch):
    """Return false when event type is invalid"""
    event = MagicMock(spec=Xlib.protocol.event.PropertyNotify)
    event.type = xwindow.Xlib.X.PropertyNotify
    event.atom = xwindow.NET_ACTIVE_WINDOW

    monkeypatch.setattr(xwindow, 'ROOT', MagicMock())
    monkeypatch.setattr(xwindow, 'DISP', MagicMock())
    result = xwindow.handle_xevent(
        event, callback=lambda *args, **kwargs: 'callback')
    assert result is True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=fixme


from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'Click>=6.0',
    'python-xlib',
    'xkbgroup',
]

test_requirements = [
    'pytest',
    'pytest-cache',
    'pytest-cov',
    'pytest-mccabe',
    'pytest-pep8',
    'pytest-pylint',
]

setup(
    name='swytcher',
    version='0.2.0',
    description="Automatically switch layout based on your active window",
    long_description=readme + '\n\n' + history,
    author="Eddie Dunn",
    author_email='eddie.dunn@gmail.com',
    url='https://github.com/eddie-dunn/swytcher',
    packages=[
        'swytcher',
    ],
    package_dir={'swytcher': 'swytcher'},
    entry_points={
        'console_scripts': [
            'swytcher=swytcher.cli:main'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='keyboard layout switcher',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



#!/usr/bin/env python
"""Get active window name and class.

Inspiration taken from http://unix.stackexchange.com/a/334293/138633
"""
import logging
from typing import Callable

import Xlib
import Xlib.display

from swytcher.util import suppress_err

log = logging.getLogger(__name__)  # pylint: disable=invalid-name

# Globals
DISP = None
ROOT = None
NET_ACTIVE_WINDOW = None
NET_WM_NAME = None
WM_NAME = None


def _set_globals():  # pragma: no cover
    """Set global X variables that are used in the module"""
    # NOTE: There should be a better way to do this, other than using globals.
    # pylint: disable=global-statement
    global DISP, ROOT, NET_ACTIVE_WINDOW, NET_WM_NAME, WM_NAME
    DISP = Xlib.display.Display()
    ROOT = DISP.screen().root
    NET_ACTIVE_WINDOW = DISP.intern_atom('_NET_ACTIVE_WINDOW')
    NET_WM_NAME = DISP.intern_atom('_NET_WM_NAME')  # UTF-8
    WM_NAME = DISP.intern_atom('WM_NAME')  # Legacy encoding
    log.debug("global vars set")


def get_window_name(window: Xlib.xobject.drawable.Window) -> str:
    """Get window name"""
    def get_name():
        """Closure"""
        log.debug("get_wm_name empty, trying other methods")
        for atom in (NET_WM_NAME, WM_NAME):
            name = window.get_full_text_property(atom)
            if name:
                return name

    name = window.get_wm_name() or get_name() or ''
    log.debug("Got window name %r", name)
    return name


@suppress_err(Exception, log, logging.ERROR, traceback=True)
@suppress_err(Xlib.error.BadWindow)  # ignore BadWindow
def handle_xevent(event: Xlib.X.PropertyNotify,
                  callback: Callable[..., None]) -> bool:
    """Handle xevent; returns true if an event was handled"""
    if event.type != Xlib.X.PropertyNotify:
        return False

    if event.atom != NET_ACTIVE_WINDOW:
        return False

    win_id = ROOT.get_full_property(NET_ACTIVE_WINDOW,
                                    Xlib.X.AnyPropertyType).value[0]

    window = DISP.create_resource_object('window', win_id)
    wmclass = window.get_wm_class() or tuple()
    wmname = get_window_name(window)
    callback(name_list=[*wmclass, wmname])
    return True


def example_handler(name_list: list):  # pragma: no cover
    """Example of a callback handler for when a window change occurs.

    This one just prints the list of strings that were found in handle_xevent.
    """
    print(name_list)


def run(callback):  # pragma: no cover
    """Runner for getting X window events.

    The callback function will be invoked with a list containing the classnames
    and window titles as a `name_list` keyword argument. If you need pass more
    parameters to the callback, use functools.partial to partially construct
    the callback function.
    """
    _set_globals()
    ROOT.change_attributes(event_mask=Xlib.X.PropertyChangeMask)
    while True:  # next_event() sleeps until we get an event
        handle_xevent(DISP.next_event(), callback)


if __name__ == '__main__':  # pragma: no cover
    run(callback=example_handler)
